# Copyright (c) 2024, Alexander Lerach
# Copyright (c) 2024, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

"""
# Analyze inline function usage

Function inlining generally is a space-time tradeoff where more inlining helps
with execution speed but increases FLASH usage. This tool helps to see which
functions are inlined where and how much FLASH usage this inlining causes.

## Command Line Interface

You can analyze the inline usage like this:

```sh
python3 -m emdbg.analyze.inline -f test.elf
```

The analysis can take some time as it has to traverse all DIEs of the DWARF data.
"""

from __future__ import annotations
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
import argparse
import logging
import posixpath

from elftools.elf.elffile import ELFFile
from elftools.dwarf.die import DIE
from elftools.dwarf.lineprogram import LineProgram
from elftools.dwarf.ranges import RangeEntry, RangeLists

import rich
from rich.console import Console
from rich.table import Table

_LOGGER = logging.getLogger(__name__)


def _rel_file_name(file_name: Path) -> str:
    """Return the filename relative to the CWD"""
    try: return str(Path(file_name).relative_to(Path().cwd()))
    except ValueError: return str(Path(file_name))

# -----------------------------------------------------------------------------
@dataclass
class InlinedInstance:
    """
    See `InlinedFunction`.
    """
    translation_unit: str
    """name of the translation unit this instance is used in."""
    file_name: Path
    """file where the function is inlined to."""
    file_line: int
    """line within the file where the function is inlined."""
    size: int
    """amount of flash used by the inlined function in this instance."""


# -----------------------------------------------------------------------------
@dataclass
class InlinedFunction:
    """
    The `InlinedFunction` represents a function that is inlined into different callers.
    It contains a list of `InlinedInstance` which represent the instance that this function is inlined into.
    """
    file_name: Path
    """file where the inline function is declared."""
    file_line: int
    """line within the file where the inline function is declared."""
    total_size: int
    """total amount of flash used due to this method being inline."""
    total_size_wo_overhead: int
    """total amount of flash used due to this method being inline, reduced by the function call instruction overheads."""
    inlined_instances: list[InlinedInstance]
    """list of instances where this function is inlined to."""

    def print(self, console: Console, num_of_instances: int, total_size: int, total_size_wo_overhead: int, call_overhead: int) -> str:
        """
        Prints the inlined function and its instances to the console in an easy-to-read format.

        :param console: console to print the output to.
        :param num_of_instances: the number of inlined instances to print. A value of `0` causes all to be displayed.
        :param total_size: total amount of FLASH used by inlined functions. Used to calculate the percentage caused by this inlined function.
        :param total_size_wo_overhead: total amount of FLASH used by inlined functions, reduced by the function call instruction overheads.
        :param call_overhead: expected instruction overhead caused by a function call.
        """
        console.print((
            f"{_rel_file_name(self.file_name)}:{self.file_line}"
            f" -- Total Size: {self.total_size}"
            f" ({(100 if total_size == 0 else (self.total_size / total_size * 100)):.2f}%)"
            f" -- Total Size without instruction overhead: {self.total_size_wo_overhead}"
            f" ({(100 if total_size_wo_overhead == 0 else (self.total_size_wo_overhead / total_size_wo_overhead * 100)):.2f}%)"
        ))

        table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
        table.add_column("Translation unit")
        table.add_column("File name")
        table.add_column("File line")
        table.add_column("Size")
        table.add_column("Call instruction overhead")
        table.add_column("%")

        for i, inlined_instance in enumerate(self.inlined_instances):
            if i < num_of_instances or num_of_instances == 0:
                table.add_row(_rel_file_name(inlined_instance.translation_unit), _rel_file_name(inlined_instance.file_name), str(inlined_instance.file_line),
                              str(inlined_instance.size), str(call_overhead),
                              "{:.2f}".format(100 if self.total_size == 0 else inlined_instance.size / self.total_size * 100))
            else:
                table.add_row("...", "...", "...")
                break

        console.print(table)


# -----------------------------------------------------------------------------
@dataclass
class AnalysisResult:
    """
    The class `AnalysisResult` represents the result of an inline analysis performed using `InlineAnalyzer`
    """
    inlined_savings_list: list[InlinedFunction]
    """inlined functions with more than one instance. These allow for potential FLASH savings."""
    inlined_no_savings_list: list[InlinedFunction]
    """inlined functions with one instance."""
    savings_total_size: int
    """overall FLASH size used by inlined functions with more than one instance."""
    no_savings_total_size: int
    """overall FLASH size used by inlined functions with one instance."""
    savings_total_size_wo_overhead: int
    """overall FLASH size used by inlined functions with more than one instance, reduced by function call instruction overheads."""
    no_savings_total_size_wo_overhead: int
    """overall FLASH size used by inlined functions with one instance, reduced by function call instruction overheads."""


# -----------------------------------------------------------------------------
class InlineAnalyzer:
    """
    Analyzes an ELF file and DWARF debugging data to identify inline functions and the instances where they are inlined to.
    This allows to identify options for a space-time tradeoff.

    :param call_overhead: expected instruction overhead caused by a function call.
    """
    def __init__(self, call_overhead: int):
        self._call_overhead = call_overhead
        self._raw_inlined_functions = defaultdict(list)


    def get_inlined_functions(self, file_name: str) -> AnalysisResult:
        """
        Returns the identified `InlinedFunction` in the given ELF file.
        This is only possible if DWARF debugging data is available which contains debug ranges and line program.

        :param file_name: path to the ELF file to analyze.

        :return: on success, a `AnalysisResult` is returned. In case of errors an exception is raised.

        :raises ValueError: if the debugging data is not sufficient of the analysis.
        """
        _LOGGER.info(f"Processing file: {file_name}")
        self._raw_inlined_functions.clear()

        with open(file_name, "rb") as f:
            elf_file = ELFFile(f)

            if not elf_file.has_dwarf_info():
                raise ValueError(f"{file_name} has no DWARF info.")

            dwarf_info = elf_file.get_dwarf_info()
            range_lists = dwarf_info.range_lists()

            for CU in dwarf_info.iter_CUs():
                line_program = dwarf_info.line_program_for_CU(CU)

                if line_program is None:
                    _LOGGER.warning("CU @ {CU.cu_offset} DWARF info is missing line program. Skipping CU.")
                    continue

                top_die = CU.get_top_DIE()
                self.die_get_inlined_rec(top_die, line_program, range_lists)
        return self.raw_inlined_to_output()


    def die_get_inlined_rec(self, die: DIE, line_program: LineProgram, range_lists: RangeLists):
        """
        Recursively traverse all DIEs of a given top DIE and extract information about inlined functions.

        :param die: DIE to be processed. Is gathered recursively after passing a top DIE.
        :param line_program: `LineProgram` extracted from the DWARF debugging data.
        :param range_lists: `RangeLists` extracted from the DWARF debugging data.
        """
        if die.tag == "DW_TAG_inlined_subroutine" and {"DW_AT_call_file"} <= die.attributes.keys():
            call_file = self.get_file_name(die.attributes["DW_AT_call_file"].value, line_program)
            call_line = die.attributes["DW_AT_call_line"].value
            size = self.get_size(die, range_lists)

            decl_die = self.resolve_die_ref(die, die.cu.cu_offset + die.attributes["DW_AT_abstract_origin"].value)

            if {"DW_AT_decl_file", "DW_AT_decl_line"} <= decl_die.attributes.keys():
                decl_file = self.get_file_name(decl_die.attributes["DW_AT_decl_file"].value, line_program)
                decl_line = decl_die.attributes["DW_AT_decl_line"].value
                translation_unit_name = self.get_translation_unit_name(die)

                called_function = InlinedInstance(translation_unit_name, call_file, call_line, size)
                self._raw_inlined_functions[(decl_file, decl_line)].append(called_function)

        # Recurse into the DIE children
        for child in die.iter_children():
                self.die_get_inlined_rec(child, line_program, range_lists)


    def get_file_name(self, file_idx: int, line_program: LineProgram) -> Path:
        """
        Returns a file name given a DIE file index. To perform this mapping the line program is required.

        :param file_idx: DIE file index for which the file name shall be returned.
        :param line_program: `LineProgram` extracted from the DWARF debugging data.

        :return: the file name for the given DIE file index. This will include the full path if the line program
                 contains the relevant data. Otherwise, only the file name without path will be returned.
        """
        lp_header = line_program.header
        file_entries = lp_header["file_entry"]
        file_entry = file_entries[file_idx - 1]
        dir_index = file_entry["dir_index"]

        if dir_index == 0:
            return Path(file_entry.name.decode())

        directory = lp_header["include_directory"][dir_index - 1]
        return Path(directory.decode()) / file_entry.name.decode()

    def get_translation_unit_name(self, die: DIE) -> str:
        """
        Returns the name of the translation unit the given DIE is contained in. If the name can't be retrieved,
        an empty string will be returned.

        :param die: DIE for which the translation unit name shall be returned.

        :return: on success, name of the translation unit. Otherwise, an empty string will be returned.
        """
        cu = self.resolve_die_ref(die, die.cu.cu_die_offset)

        if {"DW_AT_name"} <= cu.attributes.keys():
            return cu.attributes["DW_AT_name"].value.decode()
        else:
            return ""

    def get_size(self, die: DIE, range_lists: RangeLists) -> int:
        """
        Returns the size required by the given DIE. The function will try different methods to get the size depending on the
        attributes being present in the DIE. If none of the methods are successful, `0` will be returned.

        :param die: DIE for which the size shall be returned.
        :param range_lists: `RangeLists` extracted from the DWARF debugging data.

        :return: on success, the size of the DIE. Otherwise, `0` will be returned.
        """
        if {"DW_AT_high_pc"} <= die.attributes.keys():
            return die.attributes["DW_AT_high_pc"].value
        if {"DW_AT_ranges"} <= die.attributes.keys():
            if range_lists is None:
                raise ValueError(f"DWARF info is missing debug ranges, which is required for DIE: {die}.")

            range_list = range_lists.get_range_list_at_offset(die.attributes["DW_AT_ranges"].value)
            size = 0
            for entry in range_list:
                if isinstance(entry, RangeEntry):
                    size = size + (entry.end_offset - entry.begin_offset)
            return size
        return 0


    def resolve_die_ref(self, die: DIE, ref_addr: int) -> DIE:
        """
        Given a DIE containing a reference address, the DIE referenced by that address will be returned.

        :param die: DIE containing a reference address to be resolved.
        :param ref_addr: reference address pointing to another DIE.

        :return: referenced DIE.
        """
        return die.cu.get_DIE_from_refaddr(ref_addr)


    def raw_inlined_to_output(self) -> AnalysisResult:
        """
        Performs post-processing on the gathered data about inlined functions to bring it into an easy-to-use format.
        This includes wrapping the data into classes, grouping into inlined functions with and without FLASH saving potential
        and sorting by the amount of FLASH used.

        :return: a `AnalysisResult`.
        """
        inlined_savings_list = []
        inlined_no_savings_list = []

        savings_total_size = 0
        no_savings_total_size = 0
        savings_total_size_wo_overhead = 0
        no_savings_total_size_wo_overhead = 0

        for (decl_file, decl_line) in self._raw_inlined_functions:
            inlined_instances = []
            total_size = 0

            for inlined_instance in self._raw_inlined_functions[(decl_file, decl_line)]:
                inlined_instances.append(inlined_instance)
                total_size = total_size + inlined_instance.size
            total_size_wo_overhead = max(total_size - (len(inlined_instances) * self._call_overhead), 0)
            inlined_function = InlinedFunction(decl_file, decl_line, total_size, total_size_wo_overhead, inlined_instances)

            if len(inlined_instances) > 1:
                inlined_savings_list.append(inlined_function)
                savings_total_size = savings_total_size + total_size
                savings_total_size_wo_overhead = savings_total_size_wo_overhead + total_size_wo_overhead
            else:
                inlined_no_savings_list.append(inlined_function)
                no_savings_total_size = no_savings_total_size + total_size
                no_savings_total_size_wo_overhead = no_savings_total_size_wo_overhead + total_size_wo_overhead

        inlined_savings_list.sort(key=lambda x: x.total_size_wo_overhead, reverse=True)
        inlined_no_savings_list.sort(key=lambda x: x.total_size_wo_overhead, reverse=True)
        return AnalysisResult(inlined_savings_list, inlined_no_savings_list, savings_total_size, no_savings_total_size,
                              savings_total_size_wo_overhead, no_savings_total_size_wo_overhead)


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Inline function analyzer to identify potential FLASH savings.")
    parser.add_argument(
        "-f",
        dest="file",
        help="Path to the ELF file to analyze",
        required=True
    )
    parser.add_argument(
        "-n",
        help="Number of inlined functions to display. The default value of 0 causes all inlined functions to be displayed.",
        type=int,
        default=0
    )
    parser.add_argument(
        "-m",
        help="Number of inline function instances to show for each inlined function. The default value of 0 causes all instances to be displayed.",
        type=int,
        default=0
    )
    parser.add_argument(
        "--overhead",
        help="Expected instruction overhead caused by a function call, and therefore needs to be removed from the saveable space. This should include at least one branch instruction.",
        type=int,
        default=8
    )
    parser.add_argument(
        "--all",
        action="store_true",
        default=False,
        help="Include inlined functions that don't offer FLASH saving potential, i.e. functions that are only inlined once."
    )

    args = parser.parse_args()

    console = Console()
    inline_analyzer = InlineAnalyzer(args.overhead)
    analysis_result = inline_analyzer.get_inlined_functions(args.file)

    for i, inlined_function in enumerate(analysis_result.inlined_savings_list):
        if i < args.n or args.n == 0:
            inlined_function.print(console, args.m, analysis_result.savings_total_size, analysis_result.savings_total_size_wo_overhead, args.overhead)
            console.print("")
        else:
            console.print("[...]")
            break

    if args.all:
        for i, inlined_function in enumerate(analysis_result.inlined_no_savings_list):
            if i < args.n or args.n == 0:
                inlined_function.print(console, args.m, analysis_result.no_savings_total_size, analysis_result.no_savings_total_size_wo_overhead, args.overhead)
                console.print("")
            else:
                console.print("[...]")
                break

    console.print(f"Total potentially saveable space: {analysis_result.savings_total_size}")
    console.print(f"Total potentially saveable space without instruction overhead: {analysis_result.savings_total_size_wo_overhead}")

    if args.all:
        console.print(f"Total non-saveable space: {analysis_result.no_savings_total_size}")
        console.print(f"Total non-saveable space without instruction overhead: {analysis_result.no_savings_total_size_wo_overhead}")



# Copyright (c) 2024, Alexander Lerach
# Copyright (c) 2024, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from shutil import which
import argparse
import logging
import re
import subprocess

import rich
from rich.console import Console
from rich.table import Table

_LOGGER = logging.getLogger(__name__)
_BLOATY_CMD = "bloaty"


# -----------------------------------------------------------------------------
def _rel_file_name(file_path: Path) -> str:
    """Return the filename relative to the CWD."""
    try: return str(Path(file_path).relative_to(Path().cwd()))
    except ValueError: return str(Path(file_path))


# -----------------------------------------------------------------------------
def _remove_all_spaces(s: str) -> str:
    """Removes all spaces from the given string."""
    return ''.join(s.split())


# -----------------------------------------------------------------------------
def is_bloaty_installed() -> bool:
    """Checks if bloaty is installed and contained in PATH."""
    return which(_BLOATY_CMD) is not None


# -----------------------------------------------------------------------------
class SectionType(Enum):
    """
    The `SectionType` describes what kind of type a linker section is.
    """
    text = 0,
    rodata = 1,
    bss = 2,
    data = 3,
    debug = 4,
    unknown = 5


# -----------------------------------------------------------------------------
@dataclass
class Section:
    """
    The class `Section` represents a linker section.
    """
    name: str
    """name of the linker section."""
    section_type: SectionType
    """type of the linker section."""
    vm_size: int
    """VM size that the linker section uses"""
    file_size: int
    """file size that the linker section uses"""


# -----------------------------------------------------------------------------
@dataclass
class AnalysisResult:
    """
    The class `AnalysisResult` represents the result of an inline analysis performed using `SectionAnalyzer`.
    """
    file_path: Path
    """path of the file that was analyzed."""
    sections: list[Section]
    """all linker sections that were found in the analyzed file and passed the given filters."""
    total_vm_size: int
    """overall VM size that is used by all sections in the analyzed file."""
    total_file_size: int
    """overall file size that is used by all sections in the analyzed file."""

    def print(self, console: Console, overall_vm_size: int, overall_file_size: int):
        """
        Prints the analyzed file with its linker sections in an easy-to-read format.

        :param console: console to print the output to.
        :param overall_vm_size: overall VM size that is used by all sections of all analyzed files.
        :param overall_file_size: overall file size that is used by all sections of all analyzed files.
        """
        console.print((
             f"{_rel_file_name(self.file_path)} -- Total VM Size: {self.total_vm_size}"
             f" ({(100 if overall_vm_size == 0 else (self.total_vm_size / overall_vm_size * 100)):.2f}%)"
             f" -- Total File Size: {self.total_file_size}"
             f" ({(100 if overall_file_size == 0 else (self.total_file_size / overall_file_size * 100)):.2f}%)"
        ))

        table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
        table.add_column("Section")
        table.add_column("Type")
        table.add_column("VM size")
        table.add_column("File size")
        table.add_column("% (VM size)")
        table.add_column("% (File size)")

        for section in self.sections:
            table.add_row(section.name, str(section.section_type.name), str(section.vm_size), str(section.file_size),
                          "{:.2f}".format(100 if self.total_vm_size == 0 else section.vm_size / self.total_vm_size * 100),
                          "{:.2f}".format(100 if self.total_file_size == 0 else section.file_size / self.total_file_size * 100))

        console.print(table)

    def to_csv_lines(self) -> str:
        """
        Prints the analyzed file with its linker sections into the CSV format. This allows for analysis in another tool.
        """
        out = ""

        for section in self.sections:
            out = out + f"{_rel_file_name(self.file_path)},{section.name},{section.section_type.name},{section.vm_size},{section.file_size}\n"
        return out


# -----------------------------------------------------------------------------
class SectionAnalyzer:
    """
    The `SectionAnalyzer` extracts information about linker sections using bloaty.

    :param map_file_content: The contents of the map file. If `None` is passed, the output may contain sections that are removed by the linker.
    :param all_vm: Tells whether to include sections with a VM size of 0.
    :param section_filter: Regex that will be used as a filter on section names. `None` in case of no filter.
    :param type_filter: List of section types that shall be considered. Empty in case of no filters.
    """
    def __init__(self, map_file_content: str | None, all_vm: bool, section_filter: str | None, type_filter: list[str]):
        self._map_file_content = map_file_content
        self._all_vm = all_vm
        self._section_filter = section_filter
        self._type_filter = type_filter

    def get_sections(self, file_path: Path) -> AnalysisResult | None:
        """
        Analyzes the given file using bloaty and returns all identified `Section`. In case of any errors, `None` will be returned.
        The sections will only be added to the output, if they pass the given filters.
        If a map file is provided all sections that are removed by the linker will be removed from the output.

        :param file_path: the file to analyze.

        :return: on success, all identified `Section`. Otherwise, `None` will be returned.
        """
        sections = []
        total_vm_size = 0
        total_file_size = 0
        bloaty_output = self.get_bloaty_output(file_path)

        if bloaty_output:
            for line in bloaty_output.splitlines()[1:]:
                line_split = line.split(",")

                if len(line_split) == 3 and line_split[1].isdigit() and line_split[2].isdigit():
                    section_type = self.get_section_type(line_split[0])
                    section = Section(line_split[0], section_type, int(line_split[1]), int(line_split[2]))

                    if section.vm_size > 0 or self._all_vm:
                        if self.check_section_map_file(section) and self.check_section_filter(section) and self.check_section_type_filter(section):
                            sections.append(section)
                            total_vm_size = total_vm_size + section.vm_size
                            total_file_size = total_file_size + section.file_size
                else:
                    _LOGGER.error(f"bloaty output contains invalid line: {line}")
                    return None
            return AnalysisResult(file_path, sections, total_vm_size, total_file_size)
        else:
            return None

    def get_bloaty_output(self, file_path: Path) -> str | None:
        """
        Runs bloaty on the given file and collects its stdout if bloaty returns successfully. If it reports an error, `None` will be returned.
        The following args are passed to bloaty:
        - `-s vm`: Sort the output by VM size.
        - `-n 0`: Output all sections.
        - `--csv`: Output as CSV to stdout, which is easier to parse than the normal output.

        :param file_path: the file to analyze using bloaty.

        :return: on success, the bloaty stdout is returned. Otherwise, `None` will be returned.
        """
        res = subprocess.run([_BLOATY_CMD, "-s", "vm", "-n", "0", "--csv", f"{file_path.as_posix()}"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        if res.returncode == 0:
            return res.stdout.decode()
        else:
            _LOGGER.error(f"bloaty returned with: {res.returncode}. stdout: {res.stdout.decode()}. stderr: {res.stderr.decode()}")
            return None

    def get_section_type(self, section_name: str) -> SectionType:
        """Gets the type of the section based on its name. If this is not possible the type will be set to `unkown`."""
        for type in SectionType:
            if section_name.startswith(f".{type.name}"):
                return type
        return SectionType.unknown

    def check_section_map_file(self, section: Section) -> bool:
        """Check if the section is contained in the map file (and thus was not removed by the linker)."""
        return self._map_file_content is None or section.name in self._map_file_content

    def check_section_filter(self, section: Section) -> bool:
        """Check if the section name matches the user given regex. If no regex is given, always returns True."""
        return self._section_filter is None or re.compile(self._section_filter).match(section.name)

    def check_section_type_filter(self, section: Section) -> bool:
        """Check if the section type matches the user given type filter. If no filter is given, always returns True."""
        return len(self._type_filter) == 0 or section.section_type.name in self._type_filter


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Section size analyzer.")
    parser.add_argument(
        "--build-dir",
        help="Path to the build directory. If not given, the current working directory will be used instead."
    )
    parser.add_argument(
        "--map-file",
        help="Path to the map file. It is used to remove sections that got removed by the linker from the analysis results."
    )
    parser.add_argument(
        "--type-filter",
        help="Comma separated list of section types to filter for. For example '-t rodata,text' would only show rodata and text sections."
    )
    parser.add_argument(
        "--section-filter",
        help="Regex that will be applied to section names. Only the ones that match the regex will be considered."
    )
    parser.add_argument(
        "--csv",
        action="store_true",
        default=False,
        help="Write output to a CSV file instead of stdout."
    )
    parser.add_argument(
        "--all-vm",
        action="store_true",
        default=False,
        help="Also include sections with a VM size of 0."
    )

    args = parser.parse_args()

    build_dir = Path().cwd()
    map_file_content = None
    type_filter = []

    if not is_bloaty_installed():
        _LOGGER.error("bloaty is not installed. Please install it to use this tool.")
        exit(1)

    if args.build_dir:
        build_dir = Path(args.build_dir)

        if not build_dir.exists():
            _LOGGER.error(f"Given build directory: {build_dir} does not exist.")
            exit(1)

    if args.map_file:
        map_file_path = Path(args.map_file)

        if not map_file_path.exists():
            _LOGGER.error(f"Given map file: {map_file_path} does not exist.")
            exit(1)
        with open(args.map_file, "r") as f:
            map_file_content = f.read()

    if args.type_filter:
        type_filter = _remove_all_spaces(args.type_filter).split(",")

    if args.section_filter:
        try:
            re.compile(args.section_filter)
        except re.error:
            _LOGGER.error(f"Given regex: {args.section_filter} is not valid.")
            exit(1)

    console = Console()
    section_analyzer = SectionAnalyzer(map_file_content, args.all_vm, args.section_filter, type_filter)

    res = []
    overall_vm_size = 0
    overall_file_size = 0

    for file_path in build_dir.rglob("*.a"):
        analysis_result = section_analyzer.get_sections(file_path)

        if analysis_result:
            res.append(analysis_result)
            overall_vm_size = overall_vm_size + analysis_result.total_vm_size
            overall_file_size = overall_file_size + analysis_result.total_file_size

    res.sort(key=lambda x: x.total_vm_size, reverse=True)

    if args.csv:
        with open("output.csv", "w") as f:
            f.write("file_path,section,type,vm_size,file_size\n")

            for analysis_result in res:
                if analysis_result.total_vm_size > 0 or args.all_vm:
                    f.write(analysis_result.to_csv_lines())
    else:
        for analysis_result in res:
            if analysis_result.total_vm_size > 0 or args.all_vm:
                analysis_result.print(console, overall_vm_size, overall_file_size)
                console.print("")



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
import re
import math

from datetime import datetime
from itertools import zip_longest
from pathlib import Path


# -----------------------------------------------------------------------------
def gdb_getfield(value: "gdb.Value", name: str, default=None):
    """Find the field of a struct/class by name"""
    for f in value.type.fields():
        if name == f.name:
            return value[name]
    return default


def gdb_iter(obj):
    """yields the values in an array or value with a range"""
    if hasattr(obj, "value"):
        obj = obj.value()
    if hasattr(obj.type, "range"):
        for ii in range(*obj.type.range()):
            yield obj[ii]
    else:
        return []

def gdb_len(obj) -> int:
    """Computes the length of a gdb object"""
    if hasattr(obj.type, "range"):
        start, stop = obj.type.range()
        return stop - start
    else:
        return 1

def gdb_backtrace(gdb) -> str:
    """
    Unfortunately the built-in gdb command `backtrace` often crashes when
    trying to resolve function arguments whose memory is inaccessible due to
    optimizations or whose type is too complex.
    Therefore this is a simpler implementation in Python to avoid GDB crashing.

    ```
    (gdb) px4_backtrace
    #0  0x0800b3be in sched_unlock() at platforms/nuttx/NuttX/nuttx/sched/sched/sched_unlock.c:272
    #1  0x0800b59e in nxsem_post() at platforms/nuttx/NuttX/nuttx/sched/semaphore/sem_post.c:175
    #2  0x0800b5b6 in sem_post() at platforms/nuttx/NuttX/nuttx/sched/semaphore/sem_post.c:220
    #3  0x08171570 in px4::WorkQueue::SignalWorkerThread() at platforms/common/px4_work_queue/WorkQueue.cpp:151
    #4  0x08171570 in px4::WorkQueue::Add(px4::WorkItem*) at platforms/common/px4_work_queue/WorkQueue.cpp:143
    #5  0x0816f8dc in px4::WorkItem::ScheduleNow() at platforms/common/include/px4_platform_common/px4_work_queue/WorkItem.hpp:69
    #6  0x0816f8dc in uORB::SubscriptionCallbackWorkItem::call() at platforms/common/uORB/SubscriptionCallback.hpp:169
    #7  0x0816f8dc in uORB::DeviceNode::write(file*, char const*, unsigned int) at platforms/common/uORB/uORBDeviceNode.cpp:221
    #8  0x0816faca in uORB::DeviceNode::publish(orb_metadata const*, void*, void const*) at platforms/common/uORB/uORBDeviceNode.cpp:295
    #9  0x081700bc in uORB::Manager::orb_publish(orb_metadata const*, void*, void const*) at platforms/common/uORB/uORBManager.cpp:409
    #10 0x0816e6d8 in orb_publish(orb_metadata const*, orb_advert_t, void const*) at platforms/common/uORB/uORBManager.hpp:193
    #11 0x08160912 in uORB::PublicationMulti<sensor_gyro_fifo_s, (unsigned char)4>::publish(sensor_gyro_fifo_s const&) at platforms/common/uORB/PublicationMulti.hpp:92
    #12 0x08160912 in PX4Gyroscope::updateFIFO(sensor_gyro_fifo_s&) at src/lib/drivers/gyroscope/PX4Gyroscope.cpp:149
    #13 0x08040e74 in Bosch::BMI088::Gyroscope::BMI088_Gyroscope::FIFORead(unsigned long long const&, unsigned char) at src/drivers/imu/bosch/bmi088/BMI088_Gyroscope.cpp:447
    #14 0x08041102 in Bosch::BMI088::Gyroscope::BMI088_Gyroscope::RunImpl() at src/drivers/imu/bosch/bmi088/BMI088_Gyroscope.cpp:224
    #15 0x0803fb7a in I2CSPIDriver<BMI088>::Run() at platforms/common/include/px4_platform_common/i2c_spi_buses.h:343
    #16 0x0817164c in px4::WorkQueue::Run() at platforms/common/px4_work_queue/WorkQueue.cpp:187
    #17 0x08171798 in px4::WorkQueueRunner(void*) at platforms/common/px4_work_queue/WorkQueueManager.cpp:236
    #18 0x08014ccc in pthread_startup() at platforms/nuttx/NuttX/nuttx/libs/libc/pthread/pthread_create.c:59
    ```

    :return: the selected frame's backtrace without resolving function argument
    """
    frame = gdb.selected_frame()
    index = 0
    output = []
    while(frame and frame.is_valid()):
        pc = frame.pc()
        if pc > 0xffff_ff00:
            output.append(f"#{index: <2} <signal handler called>")
        else:
            line = "??"
            file = "??"
            if sal := frame.find_sal():
                line = sal.line
                if sal.symtab:
                    file = sal.symtab.fullname()
            if func := frame.function():
                func = func.print_name
                if not func.endswith(")"): func += "()"
            else:
                func = "??"
            output.append(f"#{index: <2} 0x{pc:08x} in {func} at {file}:{line}")
        frame = frame.older()
        index += 1
    return "\n".join(output)

def gdb_relative_location(gdb, location) -> str:
    """
    GDB can only place breakpoint on specific line number inside a file.
    However, if the file changes the line numbers can shift around, which makes
    this method brittle. Therefore this function finds the function inside
    the file and applies a offset or searches for a pattern inside that function
    to determine the correct line number.

    :param location:
        Location `function:+offset` or `function:regex`. In case the function
        uses static linkage you may also need to provide a unique part of the
        filename path to arbitrate multiple identically named static functions
        `file:function:+offset` or `file:function:regex`.

    :return: absolute location string `file:line_number`
    """
    parts = location.split(":")
    file_name = None
    if len(parts) == 3:
        file_name, function_name, line_pattern = parts
    elif len(parts) == 2:
        function_name, line_pattern = parts
    else:
        raise ValueError(f"Unknown location format '{location}'!")()

    function = gdb.lookup_global_symbol(function_name, gdb.SYMBOL_VAR_DOMAIN)
    if function is None:
        # Multiple static symbols may exists, we use the filename to arbitrate
        if functions := gdb.lookup_static_symbols(function_name, gdb.SYMBOL_VAR_DOMAIN):
            file_functions = [(f.symtab.fullname(), f) for f in functions]
            # Arbitrate using file name hint
            if file_name is not None:
                file_functions = [f for f in file_functions if file_name in f[0]]
            if len(file_functions) == 1:
                function = file_functions[0][1]
            else:
                raise ValueError("Multiple functions found:\n - " + "\n - ".join(functions))
    if function is None:
        raise ValueError(f"Cannot find function name '{function_name}'")
    assert function.is_function

    # Find source file and line numbers: how to use file_name?
    file = function.symtab.fullname()
    line_numbers = function.symtab.linetable().source_lines()
    lmin, lmax = min(line_numbers), max(line_numbers)

    if line_pattern.startswith("+"):
        # line offset relative to function
        line = int(line_pattern[1:]) + function.line
    else:
        # regex line pattern, read source file and find the line
        lines = Path(file).read_text().splitlines()
        lines = list(enumerate(lines[lmin:lmax]))
        for ii, line in lines:
            if re.search(line_pattern, line):
                line = lmin + ii
                break
        else:
            lines = "\n  ".join(f"{lmin+l[0]:>4}: {l[1]}" for l in lines)
            raise ValueError(f"Cannot find source line for '{line_pattern}'!\n"
                f"Function '{function_name}' stretches over lines {lmin}-{lmax}.\n")
                # f"Available source lines are:\n  {lines}")

    return f"{file}:{line}"


# -----------------------------------------------------------------------------
def _binary_search(array, value, lo: int, hi: int, direction: int):
    middle = (lo + hi) // 2
    if hi - lo <= 1: return middle

    nslice = [(lo, middle), (middle, hi)]
    pick_lower_half = array[middle] == value
    if direction > 0: pick_lower_half = 1 - pick_lower_half

    lo, hi = nslice[pick_lower_half]
    return _binary_search(array, value, lo, hi, direction)

def binary_search_last(array, value, lo: int = None, hi: int = None):
    """Binary search the last occurrance of value in an array"""
    if lo is None: lo = 0
    if hi is None: hi = len(array)
    return _binary_search(array, value, lo, hi, direction=-1)

def binary_search_first(array, value, lo: int = None, hi: int = None):
    """Binary search the first occurrance of value in an array"""
    if lo is None: lo = 0
    if hi is None: hi = len(array)
    return _binary_search(array, value, lo, hi, direction=1)


# -----------------------------------------------------------------------------
def chunks(iterable, chunk_size: int, fill=None):
    """Convert a iterable into a list of chunks and fill the rest with a value"""
    args = [iter(iterable)] * chunk_size
    return zip_longest(*args, fillvalue=fill)


def add_datetime(filename: str|Path):
    """
    Appends a filename with the current date and time:
    `Year_Month_Day_Hour_Minute_Second`

    Example: `path/name.txt` -> `path/name_2023_04_14_15_03_24.txt`
    """
    filename = Path(filename)
    return filename.with_stem(f"{filename.stem}_{datetime.now().strftime('%Y_%m_%d_%H_%M_%S')}")


def format_units(value: int | float | None, prefixes: dict[str, int] | str,
                 unit: str = None, fmt: str = None, if_zero: str = None) -> str:
    """
    Format a value with the largest prefix.

    The value is divided by the list of prefixes until it is smaller than the
    next largest prefix. Trailing zeros are replaced by spaces and padding is
    applied to align the prefixes and units. If the value is zero, the
    `if_zero` string is returned if defined.

    Predefined prefixes can be passed as a `group:input-prefix`:
    - `t`: time prefixes from nanoseconds to days.
    - `si`: SI prefixes from nano to Tera.

    .. note:: The micro prefix is µ, not u.

    Example:

    ```py
    format_units(123456, "t:µs", fmt=".1f")    # "123.5ms"
    format_units(0, "t:s", if_zero="-")        # "-"
    format_units(1234, "si:", "Hz", fmt=".2f") # "1.23kHz"
    format_units(1001, "si:", "Hz", fmt=".2f") # "1   kHz"
    format_units(2345, {"k": 1e3, "M": 1e3}, "Si", fmt=".1f") # "2.3MSi"
    ```

    :param value: An integer or floating point value. If None, an empty string is returned.
    :param prefixes:
        A dictionary of prefix string to ratio of the next largest prefix. The
        dictionary must be sorted from smallest to largest prefix. The prefix of
        the input value must be the first entry.
    :param unit: A unit string to be appended to the formatted value.
    :param fmt: A format specifier to be applied when formatting the value.
    :param if_zero: A string to be returned when the value is zero.
    """
    if value is None: return ""
    if if_zero is not None and value == 0: return if_zero

    # Find the correct prefix from a list of predefined common prefixes
    _found = False
    if prefixes.startswith("t:"):
        time_units = {"ns": 1e3, "µs": 1e3, "ms": 1e3, "s": 60, "m": 60, "h": 24, "d": 365.25/12}
        prefixes = prefixes.split(":")[1]
        prefixes = {k:v for k, v in time_units.items() if _found or (_found := (k == prefixes))}
    elif prefixes.startswith("si:"):
        prefixes = prefixes.split(":")[1]
        prefixes = {k:1e3 for k in ["n", "µ", "m", "", "k", "M", "G", "T"]
                    if _found or (_found := (k == prefixes))}

    # Divide the value until it is smaller than the next largest prefix
    for prefix, factor in prefixes.items():
        if value < factor: break
        value /= factor

    # Format the value
    value = f"{value:{fmt or ''}}"
    value_stripped = value.rstrip("0").rstrip(".")
    if if_zero is not None and value_stripped == "0": return if_zero
    # pad the value to the right to align it
    padding = max(len(p) for p in prefixes.keys()) - len(prefix)
    padding += len(value) - len(value_stripped)
    return f"{value_stripped}{padding * ' '}{prefix}{unit or ''}"


# -----------------------------------------------------------------------------
def format_table(fmtstr: str, header: list[str], rows: list[list[str]], columns: int = 1) -> str:
    """
    DEPRECATED: Use `rich.table.Table` instead!

    Formats a list of rows into a table of multiple meta-columns based on the format string.
    Example for formatting an array of registers into a table with three meta-columns:

    ```py
    fmtstr = "{:%d}  {:>%d}  {:>%d}"
    header = ["NAME", "HEX VALUE", "INT VALUE"]
    rows = [[reg, hex(value), value] for reg, value in registers.items()]
    table = utils.format_table(fmtstr, header, rows, 3)
    ```

    :param fmtstr: A string describing the spacing between columns and their
                   alignment. Must have exactly as many entries as the header a
                   rows. Example: two columns, aligned right and left with a
                   brace formatter: `"{:>%d} ({:%d})"` The `%d` is replaced by
                   the max column width by this function.
    :param header: A list of names for the header row. If you specify more than
                   one column, the header will be duplicated.
    :param rows: a list of lists of entries in the table. Each entry will be
                 converted to `str` to count the maximal length of each column.
    :param columns: If a table is very long, the header can be duplicated to the
                    right side of the table to fill more screen space.
    """
    # duplicate and join the format string for each column
    split_horizontal = columns > 0
    columns = abs(columns)
    fmtstr = "      :      ".join([fmtstr] * columns)
    fmtcnt = fmtstr.count("%d")
    column_width = [0] * fmtcnt

    # Interleave the rows for the later chunking
    fill = [""] * (fmtcnt // columns)
    if split_horizontal:
        rows = [val for tup in zip(*chunks(rows, math.ceil(len(rows) / columns), fill)) for val in tup]
    # prepend the duplicated header before the rows
    if header is not None:
        rows = [header] * columns + rows
    # Group the individual rows into multiple columns per row
    rows = [sum(rs, []) for rs in chunks(rows, columns, fill)]

    # collect each line and compute the column width
    lines = []
    for row in rows:
        if len(row) != fmtcnt:
            raise ValueError("Each row have the same number of entries as the format string")
        line = [str(l) for l in row]
        lines.append(line)
        column_width = [max(w, len(l)) for w, l in zip(column_width, line)]

    # Format the format string with the column width first
    fmtstr = fmtstr % tuple(column_width)
    # Now format the actual lines with the formatted format string
    return "\n".join(fmtstr.format(*line).rstrip() for line in lines) + "\n"


# -----------------------------------------------------------------------------
class _Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# Copyright (c) 2023, Auterion AG
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from .base import Base
from .utils import format_units
from typing import Callable, Any
import rich.box, rich.markup
from rich.table import Table
from functools import cached_property
import math


class PerfCounter(Base):
    """
    Pretty Printing Perf Counters
    """

    def __init__(self, gdb, perf_ptr: "gdb.Value"):
        super().__init__(gdb)
        self._perf = perf_ptr.cast(gdb.lookup_type("struct perf_ctr_count").pointer())
        if self.type == "PC_ELAPSED":
            self._perf = self._perf.cast(gdb.lookup_type("struct perf_ctr_elapsed").pointer())
        elif self.type == "PC_INTERVAL":
            self._perf = self._perf.cast(gdb.lookup_type("struct perf_ctr_interval").pointer())
        # print(self._perf, self.short_type, self.events, self.name)

    @cached_property
    def name(self) -> str:
        """Name of the counter"""
        try:
            return self._perf["name"].string()
        except:
            return "?"

    @cached_property
    def events(self) -> int:
        """How many events were counted"""
        return int(self._perf["event_count"])

    @cached_property
    def _types(self):
        return self._gdb.types.make_enum_dict(self._gdb.lookup_type("enum perf_counter_type"))

    @cached_property
    def type(self) -> str:
        """Counter type name"""
        for name, value in self._types.items():
            if value == self._perf["type"]:
                return name
        return "UNKNOWN"

    @cached_property
    def short_type(self) -> str:
        """The short name of the type"""
        return self.type.replace("PC_", "").capitalize()

    @cached_property
    def elapsed(self) -> int | None:
        """
        How much time has elapsed in microseconds.
        Only applies to Elapsed counters.
        """
        if self.type == "PC_ELAPSED":
            return int(self._perf["time_total"])
        return None

    @cached_property
    def first(self) -> int | None:
        """
        The first time in microseconds.
        Only applies to Interval counters.
        """
        if self.type == "PC_INTERVAL":
            return int(self._perf["time_first"])
        return None

    @cached_property
    def last(self) -> int | None:
        """
        The last time in microseconds.
        Only applies to Interval counters.
        """
        if self.type == "PC_INTERVAL":
            return int(self._perf["time_last"])
        return None

    @cached_property
    def interval(self) -> int | None:
        """
        The interval time in microseconds.
        Only applies to Interval counters.
        """
        if self.type == "PC_INTERVAL":
            return self.last - self.first
        return None

    @cached_property
    def average(self) -> int | None:
        """
        The average time in microseconds.
        Only applies to Elapsed and Interval counters.
        """
        if self.type == "PC_ELAPSED":
            return self.elapsed / self.events if self.events else 0
        elif self.type == "PC_INTERVAL":
            return (self.last - self.first) / self.events if self.events else 0
        return None

    @cached_property
    def least(self) -> int | None:
        """
        The least time in microseconds.
        Only applies to Elapsed and Interval counters.
        """
        if self.type in ["PC_ELAPSED", "PC_INTERVAL"]:
            return int(self._perf["time_least"])
        return None

    @cached_property
    def most(self) -> int | None:
        """
        The most time in microseconds.
        Only applies to Elapsed and Interval counters.
        """
        if self.type in ["PC_ELAPSED", "PC_INTERVAL"]:
            return int(self._perf["time_most"])
        return None

    @cached_property
    def rms(self) -> int | None:
        """
        The root mean square in microseconds.
        Only applies to Elapsed and Interval counters.
        """
        if self.type in ["PC_ELAPSED", "PC_INTERVAL"]:
            return 1e6 * math.sqrt(float(self._perf["M2"]) / (self.events - 1)) if self.events > 1 else 0
        return None


_PREVIOUS_COUNTERS = {}
def all_perf_counters_as_table(gdb, filter_: Callable[[PerfCounter], bool] = None,
                               sort_key: Callable[[PerfCounter], Any] = None) -> Table | None:
    """
    Pretty print all perf counters as a table. Counters that did not change
    since the last call are dimmed.

    :param filter_: A function to filter the perf counters.
    :param sort_key: A function to sort the perf counters by key.
    :returns: A rich table with all perf counters or `None` if no counters found.
    """
    if (queue := gdb.lookup_static_symbol("perf_counters")) is None:
        return None
    queue = queue.value()
    item, tail = queue["head"], queue["tail"]
    counters = []
    loop_count = 0
    while item and item != tail:
        pc = PerfCounter(gdb, item)
        counters.append(pc)
        item = item["flink"]
        loop_count += 1
        if loop_count > 1000: break
    # Filter may result in no matches
    if not counters:
        return None

    global _PREVIOUS_COUNTERS
    changed = {c for c in counters
               if (events := _PREVIOUS_COUNTERS.get(int(c._perf))) is not None
               and events != c.events}
    _PREVIOUS_COUNTERS |= {int(c._perf): c.events for c in counters}

    # Filter out the counters
    if filter_ is not None:
        counters = [c for c in counters if filter_(c)]
        if not counters:
            return None

    table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
    table.add_column("perf_ctr_count*", justify="right", no_wrap=True)
    table.add_column("Name")
    table.add_column("Events", justify="right")
    table.add_column("Elapsed", justify="right")
    table.add_column("Average", justify="right")
    table.add_column("Least", justify="right")
    table.add_column("Most", justify="right")
    table.add_column("RMS", justify="right")
    table.add_column("Interval", justify="right")
    table.add_column("First", justify="right")
    table.add_column("Last", justify="right")

    # Sort the rows by name by default and format the table
    for counter in sorted(counters, key=sort_key or (lambda p: p.name)):
        table.add_row(hex(counter._perf), rich.markup.escape(counter.name), str(counter.events),
                      format_units(counter.elapsed, "t:µs", fmt=".1f", if_zero="-"),
                      format_units(counter.average, "t:µs", fmt=".1f", if_zero="-"),
                      format_units(counter.least, "t:µs", fmt=".1f", if_zero="-"),
                      format_units(counter.most, "t:µs", fmt=".1f", if_zero="-"),
                      format_units(counter.rms, "t:µs", fmt=".3f", if_zero="-"),
                      format_units(counter.interval, "t:µs", fmt=".1f", if_zero="-"),
                      format_units(counter.first, "t:µs", fmt=".1f", if_zero="-"),
                      format_units(counter.last, "t:µs", fmt=".1f", if_zero="-"),
                      style="dim" if changed and counter not in changed else None)
    return table


